/* ====================================================================
 * Copyright (c) 2007,      Martin Hauner
 *                          http://subcommander.tigris.org
 *
 * Subcommander is licensed as described in the file doc/COPYING, which
 * you should have received as part of this distribution.
 * ====================================================================
 */

// sc
#include "config.h"
#include "ProjectWizard.h"
#include "sublib/ExternButton.h"
#include "svn/Client.h"

// qt
#include <QtGui/QLayout>
#include <QtGui/QPushButton>
#include <QtGui/QLabel>
#include <QtGui/QLineEdit>
#include <QtGui/QTextEdit>
#include <QtCore/QRegExp>
#include <QtGui/QFileDialog>

// sys
#include <cassert>


enum PageIndex
{
  IdxPage1,
  IdxPage2,
  IdxPage3,
  IdxPage4
};

ProjectWizard::ProjectWizard( QWidget* parent )
: super(parent)
{
  setCaption( _q("subcommander:new project") );

  // page 1
  {
    QWidget*     p = new QWidget(this);
    QGridLayout* l = new QGridLayout(p);

    _prjName      = new QLineEdit( _q("<project name>"), p );
    QLabel* text  = new QLabel( _q("choose a name for the project"), p );
    QLabel* label = new QLabel( _prjName, _q("Project Name:"), p );
    QLabel* info  = new QLabel( _q("A project is simply a group of related "
      "bookmarks. A bookmark can be either a repository url or a working copy "
      "path. A project can have as many bookmarks as you like.\n\n"
      "The project name can be changed at any time by choosing rename from the "
      "project (context) menu or by 'in place' editing the bookmark."), p );
    info->setWordWrap(true);

    l->setSpacing(10);
    l->addWidget( text,     0, 0, 1, 2 );
    l->addWidget( label,    1, 0 );
    l->addWidget( _prjName, 1, 1 );
    l->addWidget( info,     2, 0, 1, 2 );
    l->setRowStretch(       3, 1 );

    insertPage( p, _q("Project Name"), IdxPage1 );
    setFinishEnabled( p, true );

    connect( 
      _prjName, SIGNAL(textChanged(const QString&)),
      this,     SLOT(prjNameChanged(const QString&)) );
  }
  // page 2
  {
    QWidget*     p = new QWidget(this);
    QGridLayout* l = new QGridLayout(p);

    _wcName    = new QLineEdit( "", p );
    _wcPath    = new QLineEdit( "", p );

    QLabel* text = new QLabel( _q("To add an existing working copy to the project "
      "choose a name and enter or select its working copy path"), p );
    text->setWordWrap(true);

    QLabel* labelName = new QLabel( _prjName, _q("Name:"), p );
    QLabel* labelPath = new QLabel( _prjName, _q("Path:"), p );
    ExternButton* ext = new ExternButton(p);

    QLabel* info  = new QLabel( _q("The working copy bookmark can be set or modified at"
      " any time from the project (context) menu or the bookmark (context) menu. Its "
      "name can also be changed by 'in place' editing the bookmark.\n\n"
      "Additional working copies can be added from the project (context) menu or by drag"
      " and drop (drag in a file explorer and drop on the project bookmark)."), p );
    info->setWordWrap(true);

    l->setSpacing(10);
    l->addWidget( text,       0, 0, 1, 3 );
    l->addWidget( labelName,  1, 0 );
    l->addWidget( _wcName,    1, 1 );
    l->addWidget( labelPath,  2, 0 );
    l->addWidget( _wcPath,    2, 1 );
    l->addWidget( ext,        2, 2 );
    l->addWidget( info,       3, 0, 1, 3 );
    l->setRowStretch(         4, 1 );

    insertPage( p, _q("working copy Bookmark"), IdxPage2 );
    setFinishEnabled( p, true );

    connect( 
      _wcName, SIGNAL(textChanged(const QString&)),
      this,    SLOT(wcNameChanged(const QString&)) );
    connect( 
      _wcPath, SIGNAL(textChanged(const QString&)),
      this,    SLOT(wcPathChanged(const QString&)) );

    connect( 
      ext,  SIGNAL(clicked()),
      this, SLOT(selectPath()) );
  }
  // page3
  {
    QWidget*     p = new QWidget(this);
    QGridLayout* l = new QGridLayout(p);

    _trunkName = new QLineEdit( _q("trunk"), p );
    _trunkUrl  = new QLineEdit( "", p );

    QHBoxLayout* hl = new QHBoxLayout;
    _wc = new QPushButton( _q("get URL from working copy (%1)").arg(_q("<unset>")), p );
    hl->add(_wc);

    QLabel* text      = new QLabel( _q("enter the trunk url for your project"), p );
    QLabel* labelName = new QLabel( _prjName, _q("Name:"), p );
    QLabel* labelUrl  = new QLabel( _prjName, _q("URL:"), p );

    QLabel* info  = new QLabel( _q("The trunk url is the projects repository url. "
      "For example the trunk url of the Subcommander repository is "
      "http://subcommander.tigris.org/svn/subcommander/trunk.\n\n"
      "It is recommended to set it because it is the starting point for checking out "
      "a fresh working copy from the repository or browsing the repositoy."
      "\n\n"
      "If you do not set an url the trunk bookmark will not be visible. It can be "
      "set or modified at any time from the project (context) menu or the bookmark "
      "(context) menu. Its name can also be changed by 'in place' editing the "
      "bookmark."), p );
    info->setWordWrap(true);

    l->setSpacing(10);
    l->addWidget( text,       0, 0, 1, 2 );
    l->addWidget( labelName,  1, 0 );
    l->addWidget( _trunkName, 1, 1 );
    l->addWidget( labelUrl,   2, 0 );
    l->addWidget( _trunkUrl,  2, 1 );
    l->addLayout( hl,         3, 0, 1, 2 );
    l->addWidget( info,       4, 0, 1, 2 );
    l->setRowStretch(         5, 1 );

    insertPage( p, _q("trunk Bookmark"), IdxPage3 );
    setFinishEnabled( p, true );

    connect( 
      _trunkName, SIGNAL(textChanged(const QString&)),
      this,       SLOT(trunkNameChanged(const QString&)) );
    connect( 
      _wc,        SIGNAL(clicked()),
      this,       SLOT(trunkFromWc()) );
  }
  // page 4
  {
    QWidget*     p = new QWidget(this);
    QGridLayout* l = new QGridLayout(p);

    _branchesName  = new QLineEdit( _q("branches"), p );
    _branchesUrl   = new QLineEdit( "", p );
    _tagsName      = new QLineEdit( _q("tags"), p );
    _tagsUrl       = new QLineEdit( "", p );

    QLabel* text = new QLabel( _q("enter the branches and tags url "
      "for your project. If you don't use branches or tags just press finish."),
      p );
    text->setWordWrap(true);

    QLabel* branchesName = new QLabel( _prjName, _q("branches Name:"), p );
    QLabel* branchesUrl  = new QLabel( _prjName, _q("branches URL:"), p );
    QLabel* tagsName     = new QLabel( _prjName, _q("tags Name:"), p );
    QLabel* tagsUrl      = new QLabel( _prjName, _q("tags URL:"), p );
    _stdLayout           = new QPushButton( _q("Standard Layout"), p );

    QLabel* info  = new QLabel( _q("Setting both makes branching and tagging "
      "a lot easier because you will not have to select or enter a target url "
      "for a branch or tag when you like to create one. You will only have to "
      "give a name and subcommander will create it in the configured "
      "repository location.\n\n"
      "Use the 'Standard Layout' button to fill the urls based on the trunk url "
      "for the common trunk/branches/tags layout.\n\n"
      "If you do not set the urls the bookmarks will not be visible. Both "
      "bookmarks can be added or changed at any time from the project (context) "
      "menu or the bookmark (context) menu. Their names can also be changed by "
      "'in place' editing the corresponding bookmark."), p );
    info->setWordWrap(true);

    l->setSpacing(10);
    l->addWidget( text,          0, 0, 1, 2 );
    l->addWidget( branchesName,  1, 0 );
    l->addWidget( _branchesName, 1, 1 );
    l->addWidget( branchesUrl,   2, 0 );
    l->addWidget( _branchesUrl,  2, 1 );
    l->addWidget( tagsName,      3, 0 );
    l->addWidget( _tagsName,     3, 1 );
    l->addWidget( tagsUrl,       4, 0 );
    l->addWidget( _tagsUrl,      4, 1 );
    l->addWidget( _stdLayout,    5, 1 );
    l->addWidget( info,          6, 0, 1, 2 );
    l->setRowStretch(            7, 1 );

    insertPage( p, _q("branches/tags Bookmarks"), IdxPage4 );
    setFinishEnabled( p, true );

    connect( 
      _branchesName, SIGNAL(textChanged(const QString&)),
      this,          SLOT(branchesNameChanged(const QString&)) );
    connect( 
      _tagsName,     SIGNAL(textChanged(const QString&)),
      this,          SLOT(tagsNameChanged(const QString&)) );
    connect( 
      _stdLayout,    SIGNAL(clicked()),
      this,          SLOT(standardLayout()) );
  }

  helpButton()->hide();
}

ProjectWizard::~ProjectWizard()
{
}

void ProjectWizard::showPage( QWidget* page )
{
  super::showPage(page);

  int index = indexOf(page);

  switch(index)
  {
  case IdxPage1:
    {
      prjNameChanged( _prjName->text() );
      backButton()->hide();
      break;
    }
  case IdxPage2:
    {
      backButton()->show();
      break;
    }
  case IdxPage3:
    {
      trunkNameChanged( _trunkName->text() );

      sc::String wcPath = sc::String(_wcPath->text().utf8());
      bool enable = ! wcPath.isEmpty() && svn::Client::isWorkingCopy(wcPath);
      _wc->setEnabled(enable);

      if( enable )
        _wc->setText(_q("get URL from working copy (%1)").arg(_wcPath->text()));
      else
        _wc->setText(_q("get URL from working copy (%1)").arg(_q("<unset>")));

      nextButton()->show();
      break;
    }
  case IdxPage4:
    {
      branchesNameChanged( _branchesName->text() );
      _stdLayout->setEnabled( !_trunkUrl->text().isEmpty() );
      nextButton()->hide();
      break;
    }
  default:
    assert(false);
  }
}

void ProjectWizard::prjNameChanged(const QString& name)
{
  backButton()->setEnabled( ! name.isEmpty() );
  nextButton()->setEnabled( ! name.isEmpty() );
  finishButton()->setEnabled( ! name.isEmpty() );
}

void ProjectWizard::wcNameChanged(const QString& name)
{
  bool change =
       (! name.isEmpty() && !_wcPath->text().isEmpty())
    || (  name.isEmpty() &&  _wcPath->text().isEmpty());

  backButton()->setEnabled( change );
  nextButton()->setEnabled( change );
  finishButton()->setEnabled( change );
}

void ProjectWizard::wcPathChanged(const QString& name)
{
  bool change =
       (! name.isEmpty() && !_wcName->text().isEmpty())
    || (  name.isEmpty() &&  _wcName->text().isEmpty());

  backButton()->setEnabled( change );
  nextButton()->setEnabled( change );
  finishButton()->setEnabled( change );
}

void ProjectWizard::trunkNameChanged(const QString& name)
{
  backButton()->setEnabled( ! name.isEmpty() );
  nextButton()->setEnabled( ! name.isEmpty() );
  finishButton()->setEnabled( ! name.isEmpty() );
}

void ProjectWizard::branchesNameChanged(const QString& name)
{
  bool change = ! name.isEmpty() && !_tagsName->text().isEmpty();

  backButton()->setEnabled( change );
  nextButton()->setEnabled( change );
  finishButton()->setEnabled( change );
}

void ProjectWizard::tagsNameChanged(const QString& name)
{
  bool change = ! name.isEmpty() && !_branchesName->text().isEmpty();

  backButton()->setEnabled( change );
  nextButton()->setEnabled( change );
  finishButton()->setEnabled( change );
}

void ProjectWizard::standardLayout()
{
  QString root = _trunkUrl->text();

  root.replace( QRegExp("/trunk(/*)$"), "" );
  QString tmp;

  tmp = root + "/branches";
  _branchesUrl->setText(tmp);

  tmp = root + "/tags";
  _tagsUrl->setText(tmp);
}

void ProjectWizard::selectPath()
{
  QString res = QFileDialog::getExistingDirectory( _wcPath->text(), this,
    "get existing working copy", "subcommander:select working copy" );
  _wcPath->setText(res);
}

void ProjectWizard::trunkFromWc()
{
  sc::String wcPath = sc::String(_wcPath->text().utf8());
  sc::String wcUrl;

  svn::Client::getUrlFromPath( wcPath, wcUrl );

  _trunkUrl->setText( QString::fromUtf8(wcUrl) );
}

QString ProjectWizard::getProjectName() const
{
  return _prjName->text();
}

QString ProjectWizard::getTrunkName() const
{
  return _trunkName->text();
}

QString ProjectWizard::getTrunkUrl() const
{
  return _trunkUrl->text();
}

QString ProjectWizard::getBranchesName() const
{
  return _branchesName->text();
}

QString ProjectWizard::getBranchesUrl() const
{
  return _branchesUrl->text();
}

QString ProjectWizard::getTagsName() const
{
  return _tagsName->text();
}

QString ProjectWizard::getTagsUrl() const
{
  return _tagsUrl->text();
}

QString ProjectWizard::getWcName() const
{
  return _wcName->text();
}

QString ProjectWizard::getWcPath() const
{
  return _wcPath->text();
}
