/* ====================================================================
 * Copyright (c) 2007,      Martin Hauner
 *                          http://subcommander.tigris.org
 *
 * Subcommander is licensed as described in the file doc/COPYING, which
 * you should have received as part of this distribution.
 * ====================================================================
 */

// sc
#include "config.h"
#include "RpViewTree.h"
#include "RpViewModel.h"
#include "RpViewState.h"
#include "RpViewColumnTooltips.h"
#include "ListViewColumnTooltips.h"
#include "DirEntryLvi.h"
#include "RpSelection.h"
#include "Actions.h"
#include "Bookmark.h"
#include "DragInfo.h"
#include "SortSafe.h"
#include "DragDropMimeTypes.h"
#include "Settings.h"
#include "ListViewHeaderHandler.h"
#include "sublib/ActionStorage.h"
#include "sublib/settings/LayoutSettings.h"
#include "svn/WcStatus.h"
#include "svn/Path.h"

// qt
#include <QtCore/QTimer>
#include <QtGui/QDragMoveEvent>
#include <Qt3Support/Q3DragObject>
#include <Qt3Support/Q3PopupMenu>
#include <Qt3Support/Q3Header>


static ListViewHeaderHandler _headerHandler;


RpViewTree::RpViewTree( RpViewModel* model, ActionStorage* as, QWidget* parent )
: super(parent), _model(model), _actions(as)
{
  setName("RpViewTree");
  _state = new RpViewState( model->getBookmark() );

  // we need a timer to scroll while dragging
  QTimer* timer = new QTimer(this);
  connect( timer, SIGNAL(timeout()), this, SLOT(dragTimeout()) );
  _dragInfo = new DragInfo(timer);
  setAcceptDrops(true);
    
  new ListViewColumnTooltips(this, RpViewColumnTooltips::getTreeTooltips() );

  setRootIsDecorated(true);
  setSelectionMode( Q3ListView::Extended );
  setItemMargin( 2 );
  setShowToolTips( true );
  setAllColumnsShowFocus(true);
  //addColumn( QString::fromUtf8(_model->getRootPath()) );
  addColumn( _q("repository file/folder") );
  addColumn( _q("last rev") /*"last revision"*/ );
  addColumn( _q("size") );
  addColumn( _q("props")/*"properties"*/ );
  addColumn( _q("last author") );
  addColumn( _q("last time") );
  addColumn( "" );
  setColumnAlignment( 1, Qt::AlignRight );
  setColumnAlignment( 2, Qt::AlignRight );
  setColumnAlignment( 3, Qt::AlignRight );
  setColumnAlignment( 4, Qt::AlignRight );
  setColumnAlignment( 5, Qt::AlignRight );
  setColumnAlignment( 6, Qt::AlignLeft );
  setResizeMode(Q3ListView::LastColumn);
  setSortColumn( 0 );

  connect( _model, SIGNAL(updatedOld(const sc::String&)), this, SLOT(updateOld(const sc::String&)) );
  connect( _model, SIGNAL(updatedNew(const sc::String&)), this, SLOT(updateNew(const sc::String&)) );

  connect( this,
    SIGNAL(doubleClicked(Q3ListViewItem*,const QPoint&,int)),
    SLOT(doubleClicked(Q3ListViewItem*,const QPoint&,int)) );
  connect( this,
    SIGNAL(contextMenuRequested(Q3ListViewItem*,const QPoint&,int)),
    SLOT(contextMenuRequest(Q3ListViewItem*,const QPoint&,int)) );
  connect( this,
    SIGNAL(itemRenamed(Q3ListViewItem*,int,const QString&)),
    SLOT(itemRenamed(Q3ListViewItem*,int,const QString&)) );
  connect( this,
    SIGNAL(expanded(Q3ListViewItem*)), SLOT(expanded(Q3ListViewItem*)) );
  connect( this,
    SIGNAL(collapsed(Q3ListViewItem*)), SLOT(collapsed(Q3ListViewItem*)) );

  // we have to update the actions on each selection change so that the
  // keyboard shortcuts work.
  connect( this, SIGNAL(selectionChanged()), SLOT(selectionChanged()) );


  // setup menu
  _menu = new Q3PopupMenu(this);
  {
    QAction* action;

    action = _actions->getAction( ActionRpLog );
    action->addTo(_menu);
    action = _actions->getAction( ActionRpLogGraph );
    action->addTo(_menu);
    action = _actions->getAction( ActionRpBlame );
    action->addTo(_menu);

    _menu->insertSeparator();

    action = _actions->getAction( ActionRpBranchTag );
    action->addTo(_menu);

    _menu->insertSeparator();

    action = _actions->getAction( ActionRpCheckout );
    action->addTo(_menu);
    action = _actions->getAction( ActionRpSwitch );
    action->addTo(_menu);
    action = _actions->getAction( ActionRpDiff );
    action->addTo(_menu);
    action = _actions->getAction( ActionRpMerge );
    action->addTo(_menu);

    _menu->insertSeparator();

    action = _actions->getAction( ActionRpMkdir );
    action->addTo(_menu);
    action = _actions->getAction( ActionRpRemove );
    action->addTo(_menu);
    action = _actions->getAction( ActionRpImport );
    action->addTo(_menu);
    action = _actions->getAction( ActionRpExport );
    action->addTo(_menu);
  }

  // add the root item
  sc::String   root = _model->getBookmark()->getSource();
  DirEntryLvi* lvi  = new DirEntryLvi( this, svn::DirEntryPtr(new svn::DirEntry(root)) );
  lvi->setParentUrl(QString::fromUtf8(root));

  blockSignals(true);
  lvi->setOpen(true);
  blockSignals(false);
  
  _entries.insert( ListLvis::value_type(root,lvi) );

  Settings s;
  s.layout().getHeaderColumns(name(),header());
  _headerHandler.add(header());
}

RpViewTree::~RpViewTree()
{
  _headerHandler.del(header());

  delete _state;
  delete _dragInfo;
}

RpViewState* RpViewTree::getState()
{
  return _state;
}

void RpViewTree::contextMenuRequest( Q3ListViewItem* item, const QPoint& pos, int col )
{
  _menu->exec(pos);
}

void RpViewTree::showEvent( QShowEvent* )
{
  selectionChanged();
}

void RpViewTree::dragMoveEvent( QDragMoveEvent* e )
{
  QPoint pos = viewport()->mapFromParent(e->pos());
  Q3ListViewItem* lvi = itemAt( pos );

  if( lvi )
  {
    e->accept(lvi->acceptDrop(e));
  }
  else
  {
    e->accept(false);
  }

  _dragInfo->setDragPos(pos);
  _dragInfo->startTimer();
}

void RpViewTree::dragLeaveEvent( QDragLeaveEvent* e )
{
  _dragInfo->stopTimer();
}

void RpViewTree::dropEvent( QDropEvent* e )
{
  _dragInfo->stopTimer();

  QPoint          pos = viewport()->mapFromParent(e->pos());
  Q3ListViewItem* item = itemAt( pos );

  DirEntryLvi*   lvi = dynamic_cast<DirEntryLvi*>(item);
  sc::String     target = lvi->getEntry()->getName();

  if( ! lvi )
    return;

  if( e->action() == QDropEvent::Copy )
  {
    e->accept();
    emit copy(target);
  }
  else if( e->action() == QDropEvent::Move )
  {
    e->accept();
    emit move(target,false);
  }
}

Q3DragObject* RpViewTree::dragObject()
{
  return new Q3StoredDrag(ScMimeTypeRpFilesItem,this);
}

void RpViewTree::dragTimeout()
{
  // drag scroll
  if( visibleHeight() < contentsHeight() )
  {
    int delta = _dragInfo->getScrollDelta(contentsHeight(),visibleHeight());
    if( delta != 0 )
    {
      scrollBy( 0, delta );
      return;
    }
  }
}

void RpViewTree::itemRenamed( Q3ListViewItem* item, int col, const QString& text )
{
  emit move( svn::Path::getBaseName(sc::String(text.utf8())),true );
}

void RpViewTree::expanded( Q3ListViewItem* item )
{
  DirEntryLvi* lvi = (DirEntryLvi*)item;

  RpViewItemState state;
  state._selected = lvi->isSelected();
  state._opened   = lvi->isOpen();
  _state->setItemState( lvi->getEntry()->getName(), state );

  _model->list( lvi->getEntry()->getName(), false );
}

void RpViewTree::collapsed( Q3ListViewItem* item )
{
  DirEntryLvi* lvi = (DirEntryLvi*)item;

  RpViewItemState state;
  state._selected = lvi->isSelected();
  state._opened   = lvi->isOpen();
  _state->setItemState( lvi->getEntry()->getName(), state );
}

void RpViewTree::updateOld( const sc::String& path )
{
  svn::DirEntries entries;
  _model->getEntriesOld( path, entries );

  //DirEntryLvi* pathLvi = NULL; not needed for list..
  for( svn::DirEntries::iterator it = entries.begin(); it != entries.end(); it++ )
  {
    const sc::String& name = (*it)->getName();

    ListLvis::iterator sit = _entries.find(name);
    if( sit == _entries.end() )
    {
      assert(false); continue;
    }

    DirEntryLvi* lvi = (*sit).second;
    _entries.erase(sit);

    RpViewItemState state;
    state._selected = lvi->isSelected();
    state._opened   = lvi->isOpen();
    _state->setItemState( name, state );

    // if its the "path" item, remember it for now. It will be deleted
    // at the end. Deleting it now would delete the tree below it too.
    if( name == path )
    {
      //pathLvi = lvi;
    }
    // if it is a folder child item, take and store it for updateNew.
    else if( lvi->getEntry()->isDir() )
    {
      //if( pathLvi )
        //pathLvi->takeItem(lvi);

      _entriesup.insert( ListLvis::value_type(name,lvi) );
    }
    // drop any other item
    else
    {
      delete lvi;
    }
  }

  //delete pathLvi;
}

void RpViewTree::updateNew( const sc::String& path )
{
  SortSafe sort(this);

  svn::DirEntries entries;
  _model->getEntries( path, entries );

  for( svn::DirEntries::iterator it = entries.begin(); it != entries.end(); it++ )
  {
    svn::DirEntryPtr  entry = *it;
    const sc::String& name  = entry->getName();

    ListLvis::iterator sit = _entries.find(name);
    if( sit != _entries.end() )
    {
      // since we get each dir two times, we may already have it.
      continue;
    }

    DirEntryLvi* lvi = NULL;

    // check for a parent item..
    sc::String parent = svn::Path::getDirName(name);
    ListLvis::iterator pit = _entries.find(parent);

    if( pit == _entries.end() || name == _current /*!*/ )
    {
      // no parent, create root item
      lvi = new DirEntryLvi( this, entry );
      lvi->setParentUrl( QString::fromUtf8(_state->getPath()) );
    }
    else
    {
      // parent, create child item
      lvi = new DirEntryLvi( (*pit).second, entry );
      lvi->setParentUrl( QString::fromUtf8(parent) );
    }

    // move all childs of current items childs from their old lvi to the new lvi.
    // skip current item, ie. ".".
    if( name != path )
    {
      ListLvis::iterator sito = _entriesup.find(name);
      if( sito != _entriesup.end() )
      {
        ChildLvis childs;
        DirEntryLvi* parent = (DirEntryLvi*)((*sito).second);
        
        // temporarily add the taken item, so child methods do not crash ...
        insertItem(parent);

        DirEntryLvi* child = (DirEntryLvi*)parent->firstChild();
        while( child )
        {
          childs.push_back(child);
          child = (DirEntryLvi*)child->nextSibling();
        }

        for( ChildLvis::iterator cit = childs.begin(); cit != childs.end(); cit++ )
        {
          parent->takeItem(*cit);
          lvi->insertItem(*cit);
        }

        // ... take it again.
        takeItem(parent);
      }
    }

    if( entry->isDir() )
    {
      lvi->setExpandable(true);
    }

    const RpViewItemState& state = _state->getItemState(name);
    blockSignals(true);
    lvi->setOpen( state._opened || lvi->parent() == NULL );
    blockSignals(false);
    lvi->setSelected( state._selected );
    lvi->setVisible(true);

    if( state._selected )
    {
      setCurrentItem(lvi);
    }

    _entries.insert( ListLvis::value_type(name,lvi) );
  }

  // cleanup
  for( ListLvis::iterator it = _entriesup.begin(); it != _entriesup.end(); it++ )
  {
    delete (*it).second;
  }
  _entriesup.clear();
}

void RpViewTree::doubleClicked( Q3ListViewItem* item, const QPoint& p, int col )
{
  DirEntryLvi* lvi = dynamic_cast<DirEntryLvi*>(item);

  if( !lvi || !lvi->getEntry()->isDir() )
  {
    return;
  }

  sc::String target;

  if( lvi->text(0) == "." )
  {
    QString parent = lvi->getParentUrl();

    // already in root?
    if( parent == QString::fromUtf8(_model->getRootPath()) )
    {
      return;
    }

    int len = parent.findRev( '/' );
    parent = parent.left(len);

    // parent, leave
    target = sc::String(parent.utf8());
  }
  else
  {
    // child, enter
    target = lvi->getEntry()->getName();
  }

  _state->addPath(target);
  setCurrentDir(target);
}

void RpViewTree::setCurrentDir( const sc::String& current )
{
  QString qcurrent = QString::fromUtf8(current);
  setColumnText( 0, qcurrent );


  sc::String currentp = svn::Path::getDirName(current);

  sc::String old      = _current.isEmpty() ? _model->getBookmark()->getSource() : _current;
  sc::String oldp     = svn::Path::getDirName(old);

  DirEntryLvi* lvic  = (*_entries.find(current)).second;
  DirEntryLvi* lvio  = (*_entries.find(old)).second;
  DirEntryLvi* lviop = NULL;
  DirEntryLvi* lvicp = NULL;

  ListLvis::iterator itop = _entries.find(oldp);
  if( itop != _entries.end() )
    lviop = (*itop).second;

  ListLvis::iterator itcp = _entries.find(currentp);
  if( itcp != _entries.end() )
    lvicp = (*itcp).second;

  // restore tree.
  if( lviop )
  {
    // move item back to real parent ...
    takeItem(lvio);
    lviop->insertItem(lvio);

    // ... and restore parent path.
    lvio->setParentUrl(QString::fromUtf8(oldp));

    // old parent and up was hidden, restore visibility.
    Q3ListViewItem* item = lviop;
    while(item)
    {
      item->setVisible(true);
      item = item->parent();
    }
  }


  for( ListLvis::iterator it = _entries.begin(); it != _entries.end(); it++ )
  {
    DirEntryLvi* lvi = (*it).second;
    lvi->setVisible(true);
  }


  // set new root.
  if( lvicp )
  {
    // make new current a top level item ...
    lvicp->takeItem(lvic);
    insertItem(lvic);

    // ... and set itself as parent path.
    lvic->setParentUrl(QString::fromUtf8(current));
    //lvic->setSelected(true);
    //setCurrentItem(lvic);
    //lvic->setOpen(false);

    // hide the original parent and up.
    Q3ListViewItem* item = lvicp;
    while(item)
    {
      item->setVisible(false);
      //item->setSelectable(false);
      item = item->parent();
    }
  }

  printf( "%s -> %s\n", _current.getStr(), current.getStr() );
  _current = current;
}

void RpViewTree::refresh()
{
  setCurrentDir( _state->getPath() );
}

void RpViewTree::getSelection( svn::DirEntries& entries )
{
  Q3ListViewItemIterator it( this, Q3ListViewItemIterator::Selected );
  while( it.current() )
  {
    entries.push_back( ((DirEntryLvi*)it.current())->getEntry() );
    ++it;
  }
}

void RpViewTree::selectionChanged()
{
  // enable/disable actions based on current selection

  svn::DirEntries entries;
  getSelection(entries);

  RpSelection sel(entries);
  _model->setSelection(sel);

  _actions->enableAction( ActionCmdReload, sel.isSingle() && sel.isDir() );

  _actions->enableAction( ActionRpLog, sel.isSingle() );
  _actions->enableAction( ActionRpLogGraph, sel.isSingle() );
  _actions->enableAction( ActionRpBlame, sel.isSingle() && sel.isFile() );
  _actions->enableAction( ActionRpBranchTag, sel.isSingle() && sel.isDir() );
  _actions->enableAction( ActionRpCheckout, sel.isSingle() && sel.isDir() );
  _actions->enableAction( ActionRpSwitch, sel.isSingle() && sel.isDir() );
  _actions->enableAction( ActionRpDiff, sel.size() > 0 && sel.size() < 3 );
  _actions->enableAction( ActionRpMerge, sel.size() > 0 && sel.size() < 3 );
  _actions->enableAction( ActionRpMkdir, sel.isSingle() && sel.isDir() );
  _actions->enableAction( ActionRpRemove, !sel.isEmpty() );
}
