/* ====================================================================
 * Copyright (c) 2006-2008  Martin Hauner
 *                          http://subcommander.tigris.org
 *
 * Subcommander is licensed as described in the file doc/COPYING, which
 * you should have received as part of this distribution.
 * ====================================================================
 */

// sc
#include "config.h"
#include "WcViewFlat.h"
#include "WcViewModel.h"
#include "WcViewFlatState.h"
#include "WcViewFlatItemModel.h"
#include "WcViewFlatProxyModel.h"
#include "WcViewColumnTooltips.h"
#include "ListViewColumnTooltips.h"
#include "CursorSupport.h"
#include "WcStatusLvi.h"
#include "WcSelection.h"
#include "Actions.h"
#include "DragDropMimeTypes.h"
#include "SortSafe.h"
#include "Settings.h"
#include "ListViewHeaderHandler.h"
#include "sublib/ActionStorage.h"
#include "sublib/settings/LayoutSettings.h"
#include "svn/WcStatus.h"
#include "svn/WcEntry.h"
#include "svn/Path.h"

// qt
#include <QtCore/QTimer>
#include <QtGui/QApplication>
#include <QtGui/QAction>
#include <QtGui/QHeaderView>
#include <QtGui/QDragMoveEvent>
#include <QtGui/QItemDelegate>
#include <QtGui/QLineEdit>

#include <Qt3Support/Q3DragObject>
#include <Qt3Support/Q3PopupMenu>


static ListViewHeaderHandler _headerHandler;


/**
 * ItemDelegate for the bookmark view.
 */
class WcViewFlatItemDelegate : public QItemDelegate
{
  typedef QItemDelegate super;

public:
  WcViewFlatItemDelegate( WcViewFlat* parent ) : super(parent), _view(parent)
  {
  }

  /** Creates a text editor with frame. */
  QWidget* createEditor( QWidget* parent, const QStyleOptionViewItem& option, const QModelIndex& index ) const
  {
    QWidget* editor = super::createEditor(parent,option,index);
    editor->setStyleSheet( "border: 1px solid black" );
    return editor;
  }

  void setEditorData( QWidget* editor, const QModelIndex& index ) const
  {
    super::setEditorData(editor,index);
  }

  void setModelData( QWidget* editor, QAbstractItemModel* model, const QModelIndex& index ) const
  {
    QLineEdit* edit = static_cast<QLineEdit*>(editor);
    _view->renameItem(edit->text());
  }

  void paint( QPainter* painter, const QStyleOptionViewItem& option, const QModelIndex &index ) const
  {
    super::paint(painter,option,index);
  }

  void drawDecoration( QPainter* painter, const QStyleOptionViewItem& option, const QRect& rect, const QPixmap& pixmap ) const
  {
    super::drawDecoration(painter,option,rect,pixmap);
  }

  QSize sizeHint( const QStyleOptionViewItem& option, const QModelIndex& index ) const
  {
#ifdef Q_WS_MAC
    return super::sizeHint(option,index);
#else 
    QSize result = super::sizeHint(option,index);
    result.setHeight( result.height() + 2 * _view->style()->pixelMetric(QStyle::PM_FocusFrameVMargin) );
    return result;
#endif // Q_WS_MAC
  }

private:
  WcViewFlat* _view;
};


WcViewFlat::WcViewFlat( WcViewModel* model, ActionStorage* as, QWidget* parent )
: super(parent), _model(model), _actions(as)
{
  _itemModel  = new WcViewFlatItemModel();
  _proxyModel = new WcViewFlatProxyModel();
  _proxyModel->setSourceModel(_itemModel);
  setModel( _proxyModel );

  _state = new WcViewFlatState( model->getBookmark() );
  _proxyModel->setCurrentPath( _state->getPath() );

  setMouseTracking(true);
  setDragEnabled(true);
  setAcceptDrops(true);
  setDropIndicatorShown(true);
  setDragDropMode(DragDrop);
  setEditTriggers(SelectedClicked);
  setRootIsDecorated(false);
  setItemsExpandable(false);
  //setUniformRowHeights(true);
  //setVerticalScrollMode(ScrollPerPixel);
  //setHorizontalScrollMode(ScrollPerPixel);
  setAnimated(false);
  setSelectionMode(QAbstractItemView::ExtendedSelection);
  setSelectionBehavior(QAbstractItemView::SelectRows);
  setAllColumnsShowFocus(true);
  setTextElideMode(Qt::ElideMiddle);
  setItemDelegate( new WcViewFlatItemDelegate(this) );
  //setHighlightSections(true) !?;
  header()->setSortIndicator( 0, Qt::AscendingOrder );
  setSortingEnabled(true);
  header()->setResizeMode( QHeaderView::ResizeToContents );
  //header()->setResizeMode( 0, QHeaderView::Interactive );
  header()->setStretchLastSection(true);
  
#if 0
  header()->setStyleSheet(
    "QHeaderView {margin: 0px; border: none; padding: 0px;}"
    "QHeaderView::section {margin: 0px; border: 10px; padding: 0px;} "  // border: none;
    "QHeaderView::up-arrow {top: 0px; height: 2px; width: 2px}"
  );
#endif
  
#if 0
  new ListViewColumnTooltips( this, WcViewColumnTooltips::getFlatTooltips() );
  setItemMargin( 2 );
  setShowToolTips( true );
  setColumnAlignment( 0, Qt::AlignLeft );
  setColumnAlignment( 1, Qt::AlignCenter );
  setColumnAlignment( 2, Qt::AlignCenter );
  setColumnAlignment( 3, Qt::AlignCenter );
  setColumnAlignment( 4, Qt::AlignCenter );
  setColumnAlignment( 5, Qt::AlignCenter );
  setColumnAlignment( 6, Qt::AlignCenter );
  setColumnAlignment( 7, Qt::AlignRight );
  setColumnAlignment( 8, Qt::AlignRight );
  setColumnAlignment( 9, Qt::AlignRight );
  setResizeMode( Q3ListView::LastColumn );
  setSortColumn( 0 );
#endif
  
  connect( _model, SIGNAL(updatedOld(const sc::String&)), this, SLOT(updateOld(const sc::String&)) );
  connect( _model, SIGNAL(updatedNew(const sc::String&)), this, SLOT(updateNew(const sc::String&)) );
  connect(
    _model, SIGNAL(updatedDeepStatus(const sc::String&,const DeepStatus&)),
    this,   SLOT(updateDeepStatus(const sc::String&,const DeepStatus&)) );

  connect( _model, SIGNAL(updatedStart(const sc::String&)), this, SLOT(updateStart(const sc::String&)) );
  connect( _model, SIGNAL(updatedFinish(const sc::String&)), this, SLOT(updateFinish(const sc::String&)) );

  // setup menu
  _menu = new Q3PopupMenu(this);
  {
    QAction* action;

    action = _actions->getAction( ActionWcDiffBase );
    action->addTo(_menu);

    _menu->insertSeparator();

    action = _actions->getAction( ActionWcCommit );
    action->addTo(_menu);
    action = _actions->getAction( ActionWcLog );
    action->addTo(_menu);
    action = _actions->getAction( ActionWcLogGraph );
    action->addTo(_menu);
    action = _actions->getAction( ActionWcBlame );
    action->addTo(_menu);

    _menu->insertSeparator();

    action = _actions->getAction( ActionWcAdd );
    action->addTo(_menu);
    action = _actions->getAction( ActionWcRevert );
    action->addTo(_menu);
    action = _actions->getAction( ActionWcRemove );
    action->addTo(_menu);
    action = _actions->getAction( ActionWcUpdateRev );
    action->addTo(_menu);

    _menu->insertSeparator();

    action = _actions->getAction( ActionWcLock );
    action->addTo(_menu);
    action = _actions->getAction( ActionWcUnlock );
    action->addTo(_menu);

    _menu->insertSeparator();

    action = _actions->getAction( ActionWcProperties );
    action->addTo(_menu);
    action = _actions->getAction( ActionWcIgnore );
    action->addTo(_menu);

    _menu->insertSeparator();

    action = _actions->getAction( ActionWcEditConflict );
    action->addTo(_menu);
    action = _actions->getAction( ActionWcResolved );
    action->addTo(_menu);

    _menu->insertSeparator();

    action = _actions->getAction( ActionWcMkdir );
    action->addTo(_menu);

    action = _actions->getAction( ActionWcBranchTag );
    action->addTo(_menu);

    action = _actions->getAction( ActionWcExport );
    action->addTo(_menu);

    action = _actions->getAction( ActionWcCleanup );
    action->addTo(_menu);
  }

  Settings s;
  //s.layout().getHeaderColumns(name(),header());
  //_headerHandler.add(header());
}

WcViewFlat::~WcViewFlat()
{
  //_headerHandler.del(header());
  delete _state;
}

WcViewFlatState* WcViewFlat::getState()
{
  return _state;
}

void WcViewFlat::contextMenuEvent( QContextMenuEvent* e )
{
  _menu->exec(e->globalPos());
}

void WcViewFlat::showEvent( QShowEvent* )
{
  svn::WcStatuss statuss;
  getSelection(statuss);
  
  WcSelection sel(statuss);
  _model->setSelection(sel);

  updateMenu(sel);
}

void WcViewFlat::mousePressEvent( QMouseEvent* e )
{
  super::mousePressEvent(e);
}

void WcViewFlat::mouseMoveEvent( QMouseEvent* e )
{
  super::mouseMoveEvent(e);

  if( !(e->buttons() & Qt::LeftButton) )
    return;

  // start drag?
  if( ! dragStartable() )
    return;

  QMimeData* data = mimeData( selectedIndexes() );
  if( !data )
    return;

  // move mime data item model
  QDrag* drag = new QDrag(this);
  drag->setMimeData(data);

  drag->exec(Qt::MoveAction|Qt::CopyAction);
}

void WcViewFlat::dragEnterEvent( QDragEnterEvent* e )
{
  // we want to receive dragMoveEvents
  e->setAccepted(true);
  super::dragEnterEvent(e);
}

void WcViewFlat::dragMoveEvent( QDragMoveEvent* e )
{
  super::dragMoveEvent(e);
  e->setAccepted(false);

  if( e->mimeData()->hasFormat(ScMimeTypeWcFilesItem) )
  {
    QModelIndex dropIndex  = indexAt(e->pos());
    QModelIndex modelIndex = _proxyModel->mapToSource(dropIndex);
    if( ! modelIndex.isValid() )
      return;

    // mouse is on this item..
    svn::WcStatusPtr status = _itemModel->getStatus(modelIndex.row());

    // no dir? Then dissallow drop.
    if( ! status->isDir() )
      return;

    e->acceptProposedAction();
  }
}

void WcViewFlat::dragLeaveEvent( QDragLeaveEvent* e )
{
  super::dragLeaveEvent(e);
}

void WcViewFlat::dropEvent( QDropEvent* e )
{
  super::dropEvent(e);

  if( e->mimeData()->hasFormat(ScMimeTypeWcFilesItem) )
  {
    QModelIndex dropIndex  = indexAt(e->pos());
    QModelIndex modelIndex = _proxyModel->mapToSource(dropIndex);
    if( ! modelIndex.isValid() )
      return;

    sc::String target = _itemModel->getStatus(modelIndex.row())->getName();

    if( e->proposedAction() == Qt::CopyAction )
    {
      e->accept();
      emit copy(target);
    }
    else if( e->proposedAction() == Qt::MoveAction )
    {
      e->accept();
      emit move(target,false);
    }
  }
}

void WcViewFlat::timerEvent( QTimerEvent* e )
{
  super::timerEvent(e);
}

void WcViewFlat::renameItem( const QString& text )
{
  emit move( svn::Path::getBaseName(sc::String(text.toUtf8())), true );
}

void WcViewFlat::updateOld( const sc::String& path )
{
  // remove the old info of path and remember the selected state for the removed
  // items.

  svn::WcStatuss statuss;
  _model->getStatusOld( path, statuss );

  for( svn::WcStatuss::iterator it = statuss.begin(); it != statuss.end(); it++ )
  {
    const sc::String& name = (*it)->getName();

    // we won't get called for a deleted folder because it is no longer in the wc.
    // we have to remove the deleted folder "manually".  
    if( path != name && (*it)->isDir() && (*it)->getTextStatus() == svn::WcStatus_Deleted )
    {
      updateOld(name);
    }

    QModelIndex srcIndex = _itemModel->index(name);
    QModelIndex prxIndex = _proxyModel->mapFromSource(srcIndex);
    QModelIndex curIndex = selectionModel()->currentIndex();

    if( ! prxIndex.isValid() )
      continue;

    bool sel = selectionModel()->isRowSelected( prxIndex.row(), prxIndex.parent() );
    bool cur = curIndex.row() == prxIndex.row();

    WcViewFlatItemState state;
    state._selected = sel;
    _state->setItemState( name, state );
  }

  _itemModel->remove(path,statuss);
}

void WcViewFlat::updateNew( const sc::String& path )
{
  // add the new info of path and restore the selected state for the items.

  svn::WcStatuss statuss; bool deep;
  _model->getStatus( path, statuss, deep );

  _itemModel->add(path,statuss,deep);

  for( svn::WcStatuss::iterator it = statuss.begin(); it != statuss.end(); it++ )
  {
    const sc::String&          name  = (*it)->getName();
    const WcViewFlatItemState& state = _state->getItemState(name);

    if( ! state._selected )
      continue;

    QModelIndex srcIndex = _itemModel->index(name);
    QModelIndex prxIndex = _proxyModel->mapFromSource(srcIndex);

    if( ! prxIndex.isValid() )
      continue;

      selectionModel()->select( prxIndex, QItemSelectionModel::Rows | QItemSelectionModel::Select );
      setCurrentIndex(prxIndex);

    // we would like to scroll only if the item is not visible.
    // problem is to find out if it is not visible.
      scrollTo(prxIndex);
  }
}

// directly connect to model!?
void WcViewFlat::updateDeepStatus(const sc::String& path, const DeepStatus& status )
{
  _itemModel->setDeepStatus( path, status.modified );
}

void WcViewFlat::updateStart( const sc::String& path )
{
  //printf( "start %s\n", (const char*)path );
}

void WcViewFlat::updateFinish( const sc::String& path )
{
  //printf( "finish %s\n", (const char*)path );
}

void WcViewFlat::mouseDoubleClickEvent( QMouseEvent* e )
{
  QModelIndex dropIndex  = indexAt(e->pos());
  QModelIndex modelIndex = _proxyModel->mapToSource(dropIndex);

  svn::WcStatusPtr status = _itemModel->getStatus(modelIndex.row());
  sc::String       target = status->getName();

  if( !status->isDir() )
  {
    _model->run(target);
    return;
  }

  if( target == _state->getPath() )
  {
    // already in root?
    if( target == _model->getRootPath() )
      return;

    target = svn::Path::getDirName(status->getName());
  }

  _state->addPath(target);
  setCurrentDir(target);
}

void WcViewFlat::setCurrentDir( const sc::String& current )
{
  setWaitCursor();

  _proxyModel->setCurrentPath(current);
  emit currentChanged(current);

  restoreCursor();
}

void WcViewFlat::refresh()
{
  _proxyModel->setFilterAll(       _state->getViewAll()     );
  _proxyModel->setFilterIgnored(   _state->getViewIgnored() );
  _proxyModel->setFilterOutOfDate( _state->getViewUpdates() );

  setCurrentDir( _state->getPath() );
}

void WcViewFlat::selectionChanged( const QItemSelection& selected, const QItemSelection& deselected )
{
  svn::WcStatuss statuss;
  getSelection(statuss);
  
  WcSelection sel(statuss);
  _model->setSelection(sel);

  updateMenu(sel);
  
  // properly refresh selection changes..
  super::selectionChanged(selected,deselected);
}

void WcViewFlat::updateMenu( const WcSelection& sel )
{
  // enable/disable actions based on current selection

  _actions->enableAction( ActionCmdReload, sel.isVersionedDir() );

  _actions->enableAction( ActionWcDiffBase, sel.isDiffable() );
  _actions->enableAction( ActionWcAdd, sel.isAddable() );
  _actions->enableAction( ActionWcLog, sel.isVersioned() );
  _actions->enableAction( ActionWcRevert, sel.isRevertable() );
  _actions->enableAction( ActionWcRemove, sel.isRemoveable() );
  _actions->enableAction( ActionWcEditConflict, sel.isConflicted() );
  _actions->enableAction( ActionWcResolved, sel.isConflicted() || sel.isPropConflicted() );
  _actions->enableAction( ActionWcCommit, sel.isCommitable() );
  _actions->enableAction( ActionWcProperties, sel.isVersioned() );
  _actions->enableAction( ActionWcIgnore, sel.isUnversionedAll() );
  _actions->enableAction( ActionWcBlame, sel.isVersionedFile() );
  _actions->enableAction( ActionWcLock, sel.isVersionedFile() );
  _actions->enableAction( ActionWcUnlock, sel.isVersionedFile() );
  _actions->enableAction( ActionWcCleanup, sel.isVersionedDir() );
  _actions->enableAction( ActionWcExport, sel.isVersionedDir() );
  _actions->enableAction( ActionWcMkdir, sel.isVersionedDir() );
  _actions->enableAction( ActionWcUpdateRev, sel.isVersioned() );
  _actions->enableAction( ActionWcBranchTag, sel.isVersioned() );
  _actions->enableAction( ActionWcLogGraph, sel.isVersioned() );
}

void WcViewFlat::getSelection( svn::WcStatuss& statuss )
{
  QModelIndexList indexList = selectedIndexes();

  for( QModelIndexList::Iterator it = indexList.begin(); it != indexList.end(); it++ )
  {
    if( (*it).column() != 0 )
      continue;

    QModelIndex index = _proxyModel->mapToSource(*it);

    svn::WcStatusPtr status = _itemModel->getStatus(index.row());
    statuss.push_back(status);
  }
}

QMimeData* WcViewFlat::mimeData( const QModelIndexList& list )
{
  QMimeData *mimeData = new QMimeData();
  QByteArray encodedData;

  QDataStream stream(&encodedData, QIODevice::WriteOnly);

  foreach( QModelIndex index, list )
  {
    QModelIndex modelIndex = _proxyModel->mapToSource(index);
    if( modelIndex.isValid() )
    {
      QString value = _itemModel->data(modelIndex, Qt::DisplayRole).toString();
      stream << value;
    }
  }

  mimeData->setData( ScMimeTypeWcFilesItem, encodedData );
  return mimeData;
}
