/* ====================================================================
 * Copyright (c) 2007-2008  Martin Hauner
 *                          http://subcommander.tigris.org
 *
 * Subcommander is licensed as described in the file doc/COPYING, which
 * you should have received as part of this distribution.
 * ====================================================================
 */

// sc
#include "config.h"
#include "WcViewFlatItemModel.h"
#include "WcStatusDisplayData.h"
#include "ColorId.h"
#include "StatusId.h"
#include "DragDropMimeTypes.h"
#include "sublib/Utility.h"
#include "svn/WcStatus.h"
#include "svn/WcEntry.h"

// qt
#include <QtCore/QtAlgorithms>
#include <QtCore/QMimeData>
#include <QtCore/QList>
#include <QtGui/QPixmap>


enum Columns
{
  ColName,
  ColTextStatus,
  ColPropStatus,
  ColRpLock,
  ColWcLock,
  ColCopied,          ///< add with history
  ColSwitched,
  ColUpdate,
  ColWcRevision,
  ColCommitRevision,
  ColCommitAuthor,
  ColEmpty
};

/**
 * A helper class used by WcViewFlatItemModel to store the wc status information.
 */
class WcViewFlatItem
{
public:
  WcViewFlatItem( const svn::WcStatusPtr status )
  : _status(status), _deep(false)
  {
  }

  /** Only used for searching based on @a path. */
  WcViewFlatItem( const sc::String& path )
    : _name(path), _deep(false)
  {
  }


  /** return the wrapped status object. */
  svn::WcStatusPtr status()
  {
    return _status;
  }
  
  /** return the item name, ie. full pathname. */
  const sc::String& name() const
  {
    if(_status)
      return _status->getName();
    else
      return _name;
  }

  /** Set deep status to @a status. */
  void setDeepStatus( bool status )
  {
    _deep = status;
  }

  bool deepStatus() const
  {
    return _deep;
  }

  bool operator==( const WcViewFlatItem& other ) const
  {
    return name() == other.name();
  }

private:
  sc::String       _name;     ///< search name.

  svn::WcStatusPtr _status;   ///< the item status.
  bool             _deep;     ///< any child changed?
};


/** less than operator forWcViewFlatItems based on the filename */
bool lessThanWcViewFlatItem( const WcViewFlatItem* lhs, const WcViewFlatItem* rhs )
{
  return lhs->name() < rhs->name();
}


/** Data structure for storing the WcViewFlatItems. */
typedef QList<WcViewFlatItem*> WcViewFlatIndex;

/** This class holds the data of WcViewFlatItemModel. */ 
class WcViewFlatItemModel::Member
{
public:
  WcViewFlatItem* root;
  WcViewFlatIndex index;
};



WcViewFlatItemModel::WcViewFlatItemModel()
{
  _m = new Member();
  _m->root = new WcViewFlatItem(svn::WcStatusPtr());
}
  
WcViewFlatItemModel::~WcViewFlatItemModel()
{
  delete _m->root;
  delete _m;
}

QModelIndex WcViewFlatItemModel::index( int row, int column, const QModelIndex& parent ) const
{
  if( ! hasIndex(row,column,parent) )
    return QModelIndex();

  WcViewFlatItem* item = _m->index.at(row);
  return createIndex( row, column, item );
}

QModelIndex WcViewFlatItemModel::parent( const QModelIndex& index ) const
{
  return QModelIndex();
}

int WcViewFlatItemModel::columnCount( const QModelIndex& parent ) const
{
  return WcStatusDisplayData::columns();
}

int WcViewFlatItemModel::rowCount( const QModelIndex& parent ) const
{
  return _m->index.size();
}

QVariant WcViewFlatItemModel::headerData( int section, Qt::Orientation orientation, int role ) const
{
  if( orientation == Qt::Horizontal && role == Qt::DisplayRole )
    return WcStatusDisplayData::header(section);
  
  else if( role == Qt::TextAlignmentRole )
    return (uint)WcStatusDisplayData::alignment(section);
  
  return QVariant();
}

QVariant WcViewFlatItemModel::data( const QModelIndex& index, int role ) const
{
  if( ! index.isValid() )
    return QVariant();

  WcViewFlatItem* item = _m->index.at(index.row());
  WcStatusDisplayData data(item->status());
  
  if( role == Qt::DisplayRole || role == Qt::EditRole )
    return data.data(index.column());

  else if( role == Qt::BackgroundRole )
    return data.color();
  
  else if( role == Qt::TextAlignmentRole )
    return (uint)data.alignment(index.column());
  
  else if( role == WcStatusRole )
    return QVariant::fromValue(item->status());
  
  else if( role == DragRole )
    return data.data(0);
  
  return QVariant();

#if 0
  else if( role == Qt::DecorationRole )
    return item->standardDecoration();

  else if( role == ExpandedDecorationRole )
    return item->expandedDecoration();

  else if( role == Qt::ToolTipRole )
    return item->tooltip();

  else if( role == OnItemRole )
    return item->tooltip();

  else if( role == CurrentWcRole )
    return item->bookmark()->isCurrent();
#endif
}

bool WcViewFlatItemModel::setData( const QModelIndex& index, const QVariant& value, int role )
{
  // remove, unused
  if( ! index.isValid() || role != Qt::EditRole )
    return false;

  //WcViewFlatItem* item = _m->index.at(index.row());
  //item->setName(value.toString());

  emit dataChanged( index, index );
  return true;
}

Qt::ItemFlags WcViewFlatItemModel::flags( const QModelIndex &index ) const
{
  Qt::ItemFlags flags = 0;

  if( ! index.isValid() )
    return flags;

  flags = Qt::ItemIsEnabled | Qt::ItemIsSelectable | Qt::ItemIsDragEnabled;


  WcViewFlatItem* item = _m->index.at(index.row());

  if( index.column() == ColName )
  {
    flags |= Qt::ItemIsEditable;
  }

  if( item->status()->isDir() )
  {
    flags |= Qt::ItemIsDropEnabled;
  }

  return flags;
}

svn::WcStatusPtr WcViewFlatItemModel::getStatus( int row ) const
{
  return _m->index.at(row)->status();
}

WcViewFlatStatus WcViewFlatItemModel::getStatusX( int row ) const
{
  WcViewFlatItem* item = _m->index.at(row);
  WcViewFlatStatus result;

  result._status = item->status();
  result._deep   = item->deepStatus();

  return result;
}

QModelIndex WcViewFlatItemModel::index( const sc::String& name ) const
{
  WcViewFlatItem find(name);

  WcViewFlatIndex::Iterator it = 
    qBinaryFind( _m->index.begin(), _m->index.end(), &find, lessThanWcViewFlatItem );
    
  if( it == _m->index.end() )
    return QModelIndex();

  return createIndex( it - _m->index.begin(), 0, *it );
}

void WcViewFlatItemModel::setDeepStatus( const sc::String& path, bool deepStatus )
{
  WcViewFlatItem find(path);

  WcViewFlatIndex::Iterator it = 
    qBinaryFind( _m->index.begin(), _m->index.end(), &find, lessThanWcViewFlatItem );
    
  if( it == _m->index.end() )
    assert(false);

  WcViewFlatItem* item = *it;
  item->setDeepStatus(deepStatus);

  //int row = it - _m->index.begin();
  //QModelIndex index1 = createIndex( row, 0,                 item );
  //QModelIndex index2 = createIndex( row, item->columns()-1, item );
  //emit dataChanged( index1, index2 );
}

void WcViewFlatItemModel::add( const sc::String& path, const svn::WcStatuss& statuss, bool deepStatus )
{
  for( svn::WcStatuss::const_iterator it = statuss.begin(); it != statuss.end(); it++ )
  {
    WcViewFlatItem* item = new WcViewFlatItem(*it);

    if( it == statuss.begin() )
      item->setDeepStatus(deepStatus);
      
    WcViewFlatIndex::Iterator itf =
      qLowerBound( _m->index.begin(), _m->index.end(), item, lessThanWcViewFlatItem );
      
    int row = itf - _m->index.begin();
    
    emit beginInsertRows( QModelIndex(), row, row );
    _m->index.insert( row, item );
    emit endInsertRows();
  }
  
  //qDebug( "add: %s", (const char*)(*statuss.begin())->getName() );
  //qDebug("add");
  //print();
}

void WcViewFlatItemModel::remove( const sc::String& path, const svn::WcStatuss& statuss )
{
  for( svn::WcStatuss::const_iterator it = statuss.begin(); it != statuss.end(); it++ )
  {
    WcViewFlatItem item(*it);
    remove(item);
  }
    
  // we hit each folder twice, first as a child of its parent and a second time
  // as the root. If statuss is empty we still want to remove path if it is in
  // our internal list. So clear it just by name.
  WcViewFlatItem item(path);
  remove(item);
    
  //qDebug( "remove: %s", (const char*)(*statuss.begin())->getName() );
  //qDebug("remove");
  //print();
}

void WcViewFlatItemModel::remove(const WcViewFlatItem& item)
{
    WcViewFlatIndex::Iterator itf = 
      qBinaryFind( _m->index.begin(), _m->index.end(), &item, lessThanWcViewFlatItem );
    
    if( itf == _m->index.end() )
      return;
  
    int row = itf - _m->index.begin();

    emit beginRemoveRows( QModelIndex(), row, row );
    _m->index.erase( itf );
    emit endRemoveRows();
}

void WcViewFlatItemModel::print()
{
  int row = 0;
  for( WcViewFlatIndex::Iterator it = _m->index.begin(); it != _m->index.end(); it++ )
  {
    qDebug( "%d %s", row++, (const char*)(*it)->name() );
  }
}
