/* ====================================================================
 * Copyright (c) 2007-2008  Martin Hauner
 *                          http://subcommander.tigris.org
 *
 * Subcommander is licensed as described in the file doc/COPYING, which
 * you should have received as part of this distribution.
 * ====================================================================
 */

// sc
#include "config.h"
#include "WcViewTree.h"
#include "WcViewModel.h"
#include "WcViewFlatState.h"
#include "WcViewTreeItemModel.h"
#include "WcViewTreeProxyModel.h"
#include "WcViewColumnTooltips.h"
#include "ListViewColumnTooltips.h"
#include "WcStatusDisplayData.h"
#include "WcSelection.h"
#include "CursorSupport.h"
#include "Actions.h"
#include "Bookmark.h"
#include "DragInfo.h"
#include "DragDropMimeTypes.h"
#include "Settings.h"
#include "ListViewHeaderHandler.h"
#include "sublib/ActionStorage.h"
#include "sublib/settings/LayoutSettings.h"
#include "svn/WcStatus.h"
#include "svn/Path.h"

// qt
#include <QtCore/QTimer>
#include <QtGui/QAction>
#include <QtGui/QLineEdit>
#include <QtGui/QHeaderView>
#include <QtGui/QItemDelegate>
#include <QtGui/QDragMoveEvent>
#include <Qt3Support/Q3DragObject>
#include <Qt3Support/Q3PopupMenu>


static ListViewHeaderHandler _headerHandler;


/**
 * WcViewTree item delegate.
 */
class WcViewTreeItemDelegate : public QItemDelegate
{
  typedef QItemDelegate super;

public:
  WcViewTreeItemDelegate( WcViewTree* parent ) : super(parent), _view(parent)
  {
  }

  /** Creates a text editor with frame. */
  QWidget* createEditor( QWidget* parent, const QStyleOptionViewItem& option, const QModelIndex& index ) const
  {
    QWidget* editor = super::createEditor(parent,option,index);
    editor->setStyleSheet( "border: 1px solid black" );
    return editor;
  }

  void setEditorData( QWidget* editor, const QModelIndex& index ) const
  {
    super::setEditorData(editor,index);
  }

  void setModelData( QWidget* editor, QAbstractItemModel* model, const QModelIndex& index ) const
  {
    QLineEdit* edit = static_cast<QLineEdit*>(editor);
    _view->renameItem(edit->text());
  }

  void paint( QPainter* painter, const QStyleOptionViewItem& option, const QModelIndex &index ) const
  {
    if( index.parent() == _view->rootIndex() && !_view->isExpanded(index) )
      _view->setExpanded(index,true);

    super::paint(painter,option,index);
  }

  void drawDecoration( QPainter* painter, const QStyleOptionViewItem& option, const QRect& rect, const QPixmap& pixmap ) const
  {
    super::drawDecoration(painter,option,rect,pixmap);
  }

  QSize sizeHint( const QStyleOptionViewItem& option, const QModelIndex& index ) const
  {
#ifdef Q_WS_MAC
    return super::sizeHint(option,index);
#else 
    QSize result = super::sizeHint(option,index);
    result.setHeight( result.height() + 2 * _view->style()->pixelMetric(QStyle::PM_FocusFrameVMargin) );
    return result;
#endif // Q_WS_MAC
  }

private:
  WcViewTree* _view;
};



WcViewTree::WcViewTree( WcViewModel* model, ActionStorage* as, QWidget* parent )
: super(parent), _model(model), _actions(as)
{
  _type = /*Flat*/Tree;
  
  _itemModel  = new WcViewTreeItemModel( model->getRootPath() );
  _proxyModel = new WcViewTreeProxyModel();
  _proxyModel->setFilterModifiedFolder( (_type == Flat) ? true : false );
  _proxyModel->setSourceModel(_itemModel);
  setModel( _proxyModel );

  _state = new WcViewFlatState( model->getBookmark() );
  _proxyModel->setCurrentPath( _state->getPath() );

  setMouseTracking(true);
  setDragEnabled(true);
  setAcceptDrops(true);
  setDropIndicatorShown(true);
  setDragDropMode(DragDrop);
  setEditTriggers(SelectedClicked);
  setIndentation(10);
  setRootIsDecorated( (_type == Tree) ? true : false );
  setItemsExpandable( (_type == Tree) ? true : false );
  //setUniformRowHeights(true);
  //setVerticalScrollMode(ScrollPerPixel);
  //setHorizontalScrollMode(ScrollPerPixel);
  setAnimated(false);
  setSelectionMode(QAbstractItemView::ExtendedSelection);
  setSelectionBehavior(QAbstractItemView::SelectRows);
  setAllColumnsShowFocus(true);
  setTextElideMode(Qt::ElideMiddle);
  setItemDelegate( new WcViewTreeItemDelegate(this) );
  //setHighlightSections(true) !?;
  header()->setSortIndicator( 0, Qt::AscendingOrder );
  setSortingEnabled(true);
  header()->setResizeMode( QHeaderView::ResizeToContents );
  //header()->setResizeMode( 0, QHeaderView::Interactive );
  header()->setStretchLastSection(true);
  
#if 0
  if( _type == Flat )
  {
    setStyleSheet(
      "QTreeView::branch:open { border-image: none; image: none;} "
      "QTreeView::branch:closed { border-image: none; image: none;}"
    );
  }
#endif

#if 0
  new ListViewColumnTooltips(this, WcViewColumnTooltips::getTreeTooltips() );
#endif

  connect( _model, SIGNAL(updatedOld(const sc::String&)), this, SLOT(updateOld(const sc::String&)) );
  connect( _model, SIGNAL(updatedNew(const sc::String&)), this, SLOT(updateNew(const sc::String&)) );
  connect(
    _model, SIGNAL(updatedDeepStatus(const sc::String&,const DeepStatus&)),
    this,   SLOT(updateDeepStatus(const sc::String&,const DeepStatus&)) );

  connect( this, SIGNAL(expanded(const QModelIndex&)), this, SLOT(storeState(const QModelIndex&)) );
  connect( this, SIGNAL(collapsed(const QModelIndex&)), this, SLOT(storeState(const QModelIndex&)) );


  // setup menu
  _menu = new Q3PopupMenu(this);
  {
    QAction* action;

    action = _actions->getAction( ActionWcDiffBase );
    action->addTo(_menu);

    _menu->insertSeparator();

    action = _actions->getAction( ActionWcCommit );
    action->addTo(_menu);
    action = _actions->getAction( ActionWcLog );
    action->addTo(_menu);
    action = _actions->getAction( ActionWcLogGraph );
    action->addTo(_menu);
    action = _actions->getAction( ActionWcBlame );
    action->addTo(_menu);

    _menu->insertSeparator();

    action = _actions->getAction( ActionWcAdd );
    action->addTo(_menu);
    action = _actions->getAction( ActionWcRevert );
    action->addTo(_menu);
    action = _actions->getAction( ActionWcRemove );
    action->addTo(_menu);
    action = _actions->getAction( ActionWcUpdateRev );
    action->addTo(_menu);

    _menu->insertSeparator();

    action = _actions->getAction( ActionWcLock );
    action->addTo(_menu);
    action = _actions->getAction( ActionWcUnlock );
    action->addTo(_menu);

    _menu->insertSeparator();

    action = _actions->getAction( ActionWcProperties );
    action->addTo(_menu);
    action = _actions->getAction( ActionWcIgnore );
    action->addTo(_menu);

    _menu->insertSeparator();

    action = _actions->getAction( ActionWcEditConflict );
    action->addTo(_menu);
    action = _actions->getAction( ActionWcResolved );
    action->addTo(_menu);

    _menu->insertSeparator();

    action = _actions->getAction( ActionWcMkdir );
    action->addTo(_menu);

    action = _actions->getAction( ActionWcBranchTag );
    action->addTo(_menu);

    action = _actions->getAction( ActionWcExport );
    action->addTo(_menu);

    action = _actions->getAction( ActionWcCleanup );
    action->addTo(_menu);
  }

  Settings s;
  //s.layout().getHeaderColumns(name(),header());
  //_headerHandler.add(header());
}

WcViewTree::~WcViewTree()
{
  //_headerHandler.del(header());
  delete _state;
}

WcViewFlatState* WcViewTree::getState()
{
  return _state;
}

void WcViewTree::contextMenuEvent( QContextMenuEvent* e )
{
  _menu->exec(e->globalPos());
}

void WcViewTree::showEvent( QShowEvent* )
{
  svn::WcStatuss statuss;
  getSelection(statuss);
  
  WcSelection sel(statuss);
  _model->setSelection(sel);

  updateMenu(sel);
}

void WcViewTree::mousePressEvent( QMouseEvent* e )
{
  super::mousePressEvent(e);
}

void WcViewTree::mouseMoveEvent( QMouseEvent* e )
{
  super::mouseMoveEvent(e);

  if( !(e->buttons() & Qt::LeftButton) )
    return;

  // start drag?
  if( ! dragStartable() )
    return;

  QMimeData* data = mimeData( selectedIndexes() );
  if( !data )
    return;

  // move mime data item model
  QDrag* drag = new QDrag(this);
  drag->setMimeData(data);

  drag->exec(Qt::MoveAction|Qt::CopyAction);
}

void WcViewTree::dragEnterEvent( QDragEnterEvent* e )
{
  // we want to receive dragMoveEvents
  e->setAccepted(true);
  super::dragEnterEvent(e);
}

void WcViewTree::dragMoveEvent( QDragMoveEvent* e )
{
  super::dragMoveEvent(e);
  e->setAccepted(false);

  //printf( "action: %s\n", e->proposedAction() == Qt::MoveAction ? "move" : "copy" );

  if( e->mimeData()->hasFormat(ScMimeTypeWcFilesItem) )
  {
    QModelIndex dropIndex  = indexAt(e->pos());
    QModelIndex modelIndex = _proxyModel->mapToSource(dropIndex);
    if( ! modelIndex.isValid() )
      return;

    // mouse is on this item..
    svn::WcStatusPtr status = _itemModel->data( modelIndex, WcViewTreeItemModel::WcStatusRole ).
      value<svn::WcStatusPtr>();

    // no dir? Then dissallow drop.
    if( ! status->isDir() )
      return;

    e->acceptProposedAction();
  }
}

void WcViewTree::dragLeaveEvent( QDragLeaveEvent* e )
{
  super::dragLeaveEvent(e);
}

void WcViewTree::dropEvent( QDropEvent* e )
{
  super::dropEvent(e);

  if( e->mimeData()->hasFormat(ScMimeTypeWcFilesItem) )
  {
    QModelIndex dropIndex  = indexAt(e->pos());
    QModelIndex modelIndex = _proxyModel->mapToSource(dropIndex);
    if( ! modelIndex.isValid() )
      return;

    svn::WcStatusPtr status = _itemModel->data( modelIndex, WcViewTreeItemModel::WcStatusRole ).
      value<svn::WcStatusPtr>();

    sc::String target = status->getName();

    if( e->proposedAction() == Qt::CopyAction )
    {
      e->accept();
      emit copy(target);
    }
    else if( e->proposedAction() == Qt::MoveAction )
    {
      e->accept();
      emit move(target,false);
    }
  }
}

void WcViewTree::timerEvent( QTimerEvent* e )
{
  super::timerEvent(e);
}

void WcViewTree::renameItem( const QString& text )
{
  emit move( svn::Path::getBaseName(sc::String(text.toUtf8())), true );
}

void WcViewTree::storeState( const QModelIndex& index )
{
  WcViewFlatItemState state;
  state._selected = currentIndex() == index;
  state._opened   = isExpanded(index);

  sc::String name = index.data( WcViewTreeItemModel::NameRole ).value<sc::String>();
  _state->setItemState( name , state );
  
  //printf( "storeState: %s %s\n", state._opened ? "opened" : "closed", (const char*)name );
}

void WcViewTree::updateOld( const sc::String& path )
{
  svn::WcStatuss statuss;
  _model->getStatusOld( path, statuss );

  // remember state
  for( svn::WcStatuss::iterator it = statuss.begin(); it != statuss.end(); it++ )
  {
    const sc::String& name = (*it)->getName();

    // we won't get called for a deleted folder because it is no longer in the wc.
    // we have to remove the deleted folder "manually".
    if( path != name && (*it)->isDir() && (*it)->getTextStatus() == svn::WcStatus_Deleted )
    {
      updateOld(name);
    }

    QModelIndex srcIndex = _itemModel->index(name);
    QModelIndex prxIndex = _proxyModel->mapFromSource(srcIndex);
    QModelIndex curIndex = selectionModel()->currentIndex();

    if( ! prxIndex.isValid() )
      continue;

    bool sel  = selectionModel()->isRowSelected( prxIndex.row(), prxIndex.parent() );
    bool open = isExpanded(prxIndex);

    WcViewFlatItemState state;
    state._selected = sel;
    state._opened   = open;
    _state->setItemState( name, state );
  }

  _itemModel->remove(path,statuss);
}

void WcViewTree::updateNew( const sc::String& path )
{
  svn::WcStatuss statuss; bool deep;
  _model->getStatus( path, statuss, deep );

  _itemModel->add(path,statuss,deep);

  for( svn::WcStatuss::iterator it = statuss.begin(); it != statuss.end(); it++ )
  {
    const sc::String&          name  = (*it)->getName();
    const WcViewFlatItemState& state = _state->getItemState(name);

    QModelIndex srcIndex = _itemModel->index(name);
    QModelIndex prxIndex = _proxyModel->mapFromSource(srcIndex);
    
    if(!prxIndex.isValid())
      continue;
    
    if(state._selected)
    {
      selectionModel()->select(
        prxIndex, QItemSelectionModel::Rows | QItemSelectionModel::Select );
      setCurrentIndex(prxIndex);
      
      // we would like to scroll only if the item is not visible.
      // problem is to find out if it is not visible...
      scrollTo(prxIndex);
    }
      
    if(state._opened)
      setExpanded( prxIndex, true );
  }
}


static sc::String breakpoint("/Users/hauner/Development/Subcommander/sc/po");

void WcViewTree::setChildsExpanded( const QModelIndex& index )
{
  if(_proxyModel->hasChildren(index))
  {
    int numRows = _proxyModel->rowCount(index);
    
    for( int row = 0; row < numRows; ++row )
    {
      QModelIndex idx = index.child(row,0);
      
      svn::WcStatusPtr status =
      idx.data( WcViewTreeItemModel::WcStatusRole ).value<svn::WcStatusPtr>();
      const sc::String& name = status->getName();
      
      if(name==breakpoint)
        int i = 0;
      
      const WcViewFlatItemState& state = _state->getItemState(name);
      
      if( status->isDir() && !isExpanded(idx) )
        setExpanded(idx,true);
    }
  }
}

void WcViewTree::updateDeepStatus(const sc::String& path, const DeepStatus& status )
{
  if( _type == Flat && status.modified )
  {
    if(path==breakpoint)
      int i= 0;
    
    QModelIndex srcIndex = _itemModel->index(path);
    QModelIndex prxIndex = _proxyModel->mapFromSource(srcIndex);
    
    if(prxIndex.isValid())
    {
      setChildsExpanded(prxIndex);

      if( !isExpanded(prxIndex) )
        setExpanded( prxIndex, true );          
    }
    else
      printf("updateDeepStatus - %s\n", (const char*)path );
  }
  
  _itemModel->setDeepStatus( path, status.modified );
}

void WcViewTree::mouseDoubleClickEvent( QMouseEvent* e )
{
  QModelIndex clickIndex = indexAt(e->pos());

  if( ! clickIndex.isValid() )
    assert(false);
  
  svn::WcStatusPtr status = clickIndex.data(WcViewTreeItemModel::WcStatusRole).value<svn::WcStatusPtr>();
  sc::String target = status->getName();

  if( !status->isDir() )
    return;

  // ignore clicks on ourself
  if( target == _state->getPath() )
    return; 

  _state->addPath(target);
  setCurrentDir(target);
}

void WcViewTree::setCurrentDir( const sc::String& current )
{
  setWaitCursor();

  // set new root index.
  QModelIndex modelIndex = _itemModel->index(current);
  QModelIndex proxyIndex = _proxyModel->mapFromSource(modelIndex);
  QModelIndex parentIndex = proxyIndex.parent();
  setRootIndex( parentIndex );
  setExpanded( parentIndex, true );
  
  _proxyModel->setCurrentPath(current);
  emit currentChanged(current);

  // re open folder items.
  QModelIndex modelIndex2 = _itemModel->index(current);
  QModelIndex proxyIndex2 = _proxyModel->mapFromSource(modelIndex2);
  restoreState( proxyIndex2 );  
  
  restoreCursor();
}

void WcViewTree::restoreState( const QModelIndex& index )
{
  if(_proxyModel->hasChildren(index))
  {
    int numRows = _proxyModel->rowCount(index);
    
    for( int row = 0; row < numRows; ++row )
    {
      QModelIndex idx = index.child(row,0);

      svn::WcStatusPtr status =
        idx.data( WcViewTreeItemModel::WcStatusRole ).value<svn::WcStatusPtr>();
      const sc::String& name = status->getName();
      
      const WcViewFlatItemState& state = _state->getItemState(name);
      if( status->isDir() && state._opened )
      {
        if(!isExpanded(idx))
          setExpanded(idx,true);
        
        restoreState(idx);
      }
    }
  }
}
  
void WcViewTree::refresh()
{
  _proxyModel->setFilterAll(       _state->getViewAll()     );
  _proxyModel->setFilterIgnored(   _state->getViewIgnored() );
  _proxyModel->setFilterOutOfDate( _state->getViewUpdates() );

  setCurrentDir( _state->getPath() );
}

void WcViewTree::selectionChanged( const QItemSelection& selected, const QItemSelection& deselected )
{
  svn::WcStatuss statuss;
  getSelection(statuss);
  
  WcSelection sel(statuss);
  _model->setSelection(sel);

  updateMenu(sel);

  // properly refresh selection changes..
  super::selectionChanged(selected,deselected);
}

void WcViewTree::updateMenu( const WcSelection& sel )
{
  // enable/disable actions based on current selection

  _actions->enableAction( ActionCmdReload, sel.isVersionedDir() );

  _actions->enableAction( ActionWcDiffBase, sel.isDiffable() );
  _actions->enableAction( ActionWcAdd, sel.isAddable() );
  _actions->enableAction( ActionWcLog, sel.isVersioned() );
  _actions->enableAction( ActionWcRevert, sel.isRevertable() );
  _actions->enableAction( ActionWcRemove, sel.isRemoveable() );
  _actions->enableAction( ActionWcEditConflict, sel.isConflicted() );
  _actions->enableAction( ActionWcResolved, sel.isConflicted() || sel.isPropConflicted() );
  _actions->enableAction( ActionWcCommit, sel.isCommitable() );
  _actions->enableAction( ActionWcProperties, sel.isVersioned() );
  _actions->enableAction( ActionWcIgnore, sel.isUnversionedAll() );
  _actions->enableAction( ActionWcBlame, sel.isVersionedFile() );
  _actions->enableAction( ActionWcLock, sel.isVersionedFile() );
  _actions->enableAction( ActionWcUnlock, sel.isVersionedFile() );
  _actions->enableAction( ActionWcCleanup, sel.isVersionedDir() );
  _actions->enableAction( ActionWcExport, sel.isVersionedDir() );
  _actions->enableAction( ActionWcMkdir, sel.isVersionedDir() );
  _actions->enableAction( ActionWcUpdateRev, sel.isVersioned() );
  _actions->enableAction( ActionWcBranchTag, sel.isVersioned() );
  _actions->enableAction( ActionWcLogGraph, sel.isVersioned() );
}

void WcViewTree::getSelection( svn::WcStatuss& statuss )
{
  QModelIndexList indexList = selectedIndexes();

  for( QModelIndexList::Iterator it = indexList.begin(); it != indexList.end(); it++ )
  {
    QModelIndex index = _proxyModel->mapToSource(*it);

    if( index.column() != 0 )
      continue;

    svn::WcStatusPtr status = _itemModel->data( index, WcViewTreeItemModel::WcStatusRole ).
      value<svn::WcStatusPtr>();

    statuss.push_back(status);
  }
}

QMimeData* WcViewTree::mimeData( const QModelIndexList& list )
{
  QMimeData *mimeData = new QMimeData();
  QByteArray encodedData;

  QDataStream stream(&encodedData, QIODevice::WriteOnly);

  foreach( QModelIndex index, list )
  {
    QModelIndex modelIndex = _proxyModel->mapToSource(index);
    if( modelIndex.isValid() )
    {
      QString value = _itemModel->data(modelIndex, WcViewTreeItemModel::DragRole).toString();
      stream << value;
    }
  }

  mimeData->setData( ScMimeTypeWcFilesItem, encodedData );
  return mimeData;
}
