/* ====================================================================
 * Copyright (c) 2007-2008  Martin Hauner
 *                          http://subcommander.tigris.org
 *
 * Subcommander is licensed as described in the file doc/COPYING, which
 * you should have received as part of this distribution.
 * ====================================================================
 */

// sc
#include "config.h"
#include "WcViewTreeItemModel.h"
#include "DragDropMimeTypes.h"
#include "WcStatusDisplayData.h"
#include "ColorId.h"
#include "sublib/QSupportAlgorithms.h"
#include "sublib/Utility.h"
#include "svn/WcStatus.h"
#include "svn/WcEntry.h"
#include "svn/Path.h"

// qt
#include <QtCore/QtAlgorithms>
#include <QtCore/QMimeData>
#include <QtCore/QList>
#include <QtGui/QPixmap>



class WcViewTreeItem;

/** the child item list of a @a WcViewTreeItem. */
typedef QList<WcViewTreeItem*> WcViewTreeItems;

/** WcViewTreeItems iterator. */
typedef QListIterator<WcViewTreeItem*> WcViewTreeItemsIterator;


/** Child item. */
class WcViewTreeItem
{
public:
  WcViewTreeItem( svn::WcStatusPtr status ) : _status(status)
  {
  }

  /** Get the items @a svn::WcStatusPtr. */
  svn::WcStatusPtr status() const
  {
    return _status;
  }

  /** Get path. */
  sc::String path() const
  {
    return _status->getName();
  }

private:
  svn::WcStatusPtr _status;
};



/**
 * Entry in @a WcViewTreeItemIndex. Holds @a WcViewTreeItems of a single paths
 * direct children.
 */
class WcViewTreeItemFolder
{
  /** compare class. */
  class equal
  {
  public:
    bool operator()( WcViewTreeItem* item, const sc::String& path )
    {
      return item->path() < path;
    }

    bool operator()( const sc::String& path, WcViewTreeItem* item )
    {
      return path < item->path();
    }
  };

public:
  /** construct entry for @a path, its children @a statuss with deep state @a deep. */
  WcViewTreeItemFolder( const sc::String& path, const svn::WcStatuss& statuss, bool deep )
    : _path(path), _deep(deep)
  {
    if( statuss.size() <= 1 )
      return;

    for( svn::WcStatuss::const_iterator it = ++statuss.begin(); it != statuss.end(); it++ )
    {
      _childs.append( new WcViewTreeItem(*it) );
    }
  }

  ~WcViewTreeItemFolder()
  {
    for( int i = _childs.size()-1; i >= 0; --i )
    {
      delete _childs.takeAt(i);
    }
  }

  /** Get this items path. */
  const sc::String& path() const
  {
    return _path;
  }

  /** Get the deep status. */
  bool deepStatus() const
  {
    return _deep;
  }

  /* Set the deep status to @a deep. */
  void setDeepStatus( bool deep )
  {
    _deep = deep;
  }

  /** Get the child item at row @a row. */
  WcViewTreeItem* child( int row ) const
  {
    return _childs[row];
  }

  /** Get the number of child items. */
  int childCount() const
  {
    return _childs.size();
  }

  /** Get row index for @a child. Returns -1 if @a child is not a child. */
  int row( const sc::String& child ) const
  {
    WcViewTreeItems::const_iterator it = sc::binaryFind( _childs.begin(), _childs.end(), child, equal() );

    if( it == _childs.end() )
      return -1;

    return it - _childs.begin();
  }

  /** Get row index for @a child. Returns -1 if @a child is not a child. */ 
  int row( WcViewTreeItem* child ) const
  {
    return _childs.indexOf(child);
  }

  /** Returns the row at which @a child should be added. */
  int insertRow( const sc::String& child )
  {
    return sc::lowerBound( _childs.begin(), _childs.end(), child, equal() ) - _childs.begin();
  }

  /** Insert @a item at @a row. */
  void insert( int row, WcViewTreeItem* item )
  {
    _childs.insert( row, item );
  }

  /** Remove @a row. */
  void remove( int row )
  {
    delete _childs.takeAt(row);
  }

private:
  sc::String      _path;
  bool            _deep;

  WcViewTreeItems _childs;
};



/** The main index. */
typedef QMap< sc::String, WcViewTreeItemFolder* > WcViewTreeItemIndex;


/** WcViewTreeItemModel member class. */
class WcViewTreeItemModelMember
{
public:
  /** Construct a WcViewTreeItemModelMember with the provided @a root path. */
  WcViewTreeItemModelMember( const sc::String& root )
    : _root(parent(root),svn::WcStatuss(),false)
  {
  }

  /** Extract a WcViewTreeItem* from @a index. */
  WcViewTreeItem* getItem( const QModelIndex& index )
  {
    return static_cast<WcViewTreeItem*>(index.internalPointer());
  }

  /** Get the WcViewTreeItemFolder for @a path. Returns NULL if @a path is not
   * available. */
  WcViewTreeItemFolder* getItemFolder( const sc::String& path )
  {
    if( path == _root.path() )
      return &_root;

    WcViewTreeItemIndex::iterator it = _index.find(path);
    if( it == _index.end() )
      return NULL;

    return *it;
  }

  /** Get the parent WcViewTreeItemFolder for @a path. Returns NULL if @a path is
   * the root path or if the parent path is not available. */
  WcViewTreeItemFolder* getParentFolder( const sc::String& path )
  {
    // the invisible root has no parent
    if( path == _root.path() )
      return NULL;

    sc::String p = parent(path);

    if( p == _root.path() )
      return &_root;

    WcViewTreeItemIndex::iterator it = _index.find(p);
    if( it == _index.end() )
      return NULL;

    return *it;
  }

  /** Find the WcViewTreeItemFolder for @a path. Returns NULL if @ path is not a
   * folder. */
  WcViewTreeItemFolder* findFolder( const sc::String& path )
  {
    WcViewTreeItemIndex::iterator it = _index.find(path);
    if( it == _index.end() )
      return NULL;

    return *it;
  }

  /** Get parent path. */
  sc::String parent( const sc::String& path ) const
  {
    return svn::Path::getDirName(path);
  }

  WcViewTreeItemIndex& index()
  {
    return _index;
  }

  WcViewTreeItemFolder* root()
  {
    return &_root;
  }

private:
  WcViewTreeItemFolder _root;
  WcViewTreeItemIndex  _index;
};


///////////////////////////////////////////////////////////////////////////////


WcViewTreeItemModel::WcViewTreeItemModel( const sc::String& root )
{
  _m = new WcViewTreeItemModelMember(root);
}
  
WcViewTreeItemModel::~WcViewTreeItemModel()
{
  delete _m;
}

QModelIndex WcViewTreeItemModel::index( int row, int column, const QModelIndex& parent ) const
{
  if( ! hasIndex(row,column,parent) )
    return QModelIndex();

  WcViewTreeItem* childItem = NULL;
  if( ! parent.isValid() )
  {
    WcViewTreeItemFolder* parentFolder = _m->root();
    childItem = parentFolder->child(row);
  }
  else
  {
    WcViewTreeItem* parentItem = _m->getItem(parent);
    WcViewTreeItemFolder* parentFolder = _m->findFolder(parentItem->path());
    childItem = parentFolder->child(row);
  }

  assert(childItem);
  return createIndex( row, column, childItem );
}

QModelIndex WcViewTreeItemModel::parent( const QModelIndex& index ) const
{
  // index is invisible root?
  if( !index.isValid() )
    return QModelIndex();

  WcViewTreeItem*       item   = _m->getItem(index);
  WcViewTreeItemFolder* parent = _m->getParentFolder(item->path());

  // parent is invisible root?
  if( parent->path() == _m->root()->path() )
    return QModelIndex();

  WcViewTreeItemFolder* grandparent = _m->getParentFolder(parent->path());

  int row = grandparent->row(parent->path());
  return createIndex( row, 0, grandparent->child(row) );
}

int WcViewTreeItemModel::columnCount( const QModelIndex& parent ) const
{
  return WcStatusDisplayData::columns();
}

int WcViewTreeItemModel::rowCount( const QModelIndex& parent ) const
{
  if( ! parent.isValid() )
    return _m->root()->childCount();

  WcViewTreeItem* item = _m->getItem(parent);
  assert(item);

  WcViewTreeItemFolder* folder = _m->findFolder(item->path());

  if( folder )
    return folder->childCount();
  else
    return 0;
}

QVariant WcViewTreeItemModel::headerData( int section, Qt::Orientation orientation, int role ) const
{
  if( orientation == Qt::Horizontal && role == Qt::DisplayRole )
    return WcStatusDisplayData::header(section);
  
  else if( role == Qt::TextAlignmentRole )
    return (uint)WcStatusDisplayData::alignment(section);

  return QVariant();
}

QVariant WcViewTreeItemModel::data( const QModelIndex& index, int role ) const
{
  if( ! index.isValid() )
    return QVariant();

  WcViewTreeItem* item = _m->getItem(index);
  WcStatusDisplayData data(item->status());

  if( role == Qt::DisplayRole || role == Qt::EditRole )
  {
    if( index.column() == WcStatusDisplayData::ColDeepStatus && item->status()->isDir() )
    {
      WcViewTreeItemFolder* folder = _m->getItemFolder(item->path());

      if( folder && folder->deepStatus() )
        return "*";
    }
    return data.data(index.column());
  }
  else if( role == Qt::BackgroundRole )
    return data.color();

  else if( role == Qt::TextAlignmentRole )
    return (uint)data.alignment(index.column());

  else if( role == WcStatusRole )
    return QVariant::fromValue(item->status());

  else if( role == DeepRole )
  {
    if( ! item->status()->isDir() )
      return QVariant(false);
    
    WcViewTreeItemFolder* folder = _m->getItemFolder(item->path());
    if(folder)
      return QVariant(folder->deepStatus());
    else
      return QVariant(false);
  }
  else if( role == NameRole )
    return QVariant::fromValue(item->path());

  else if( role == DragRole )
    return data.data(0);

  return QVariant();
  
#if 0
  else if( role == Qt::DecorationRole )
    return item->standardDecoration();

  else if( role == ExpandedDecorationRole )
    return item->expandedDecoration();

  else if( role == Qt::ToolTipRole )
    return item->tooltip();

  else if( role == OnItemRole )
    return item->tooltip();

  else if( role == CurrentWcRole )
    return item->bookmark()->isCurrent();
#endif
}

bool WcViewTreeItemModel::setData( const QModelIndex& index, const QVariant& value, int role )
{
  // remove unused.
  return false;
}

Qt::ItemFlags WcViewTreeItemModel::flags( const QModelIndex &index ) const
{
  Qt::ItemFlags flags;

  if( ! index.isValid() )
    return flags;

  flags = Qt::ItemIsEnabled | Qt::ItemIsSelectable | Qt::ItemIsDragEnabled;

  WcViewTreeItem* item = _m->getItem(index);

  if( index.column() == WcStatusDisplayData::ColName )
  {
    flags |= Qt::ItemIsEditable;
  }

  if( item->status()->isDir() )
  {
    flags |= Qt::ItemIsDropEnabled;
  }

  return flags;
}

QModelIndex WcViewTreeItemModel::index( const sc::String& path ) const
{
  WcViewTreeItemFolder* parentFolder = _m->getParentFolder(path);
  assert(parentFolder);
  
  int row = parentFolder->row(path);
  if( row == -1 )
    return QModelIndex();

  WcViewTreeItem* child = parentFolder->child(row);
  assert(child);

  return createIndex( row, 0, child );
}

void WcViewTreeItemModel::setDeepStatus( const sc::String& path, bool deepStatus )
{
  WcViewTreeItemFolder* item = _m->getItemFolder(path);
  if( !item )
    assert(false);

  if( item->deepStatus() == deepStatus )
    return;
  
  item->setDeepStatus(deepStatus);

  // \todo required ???
  layoutAboutToBeChanged();
  QModelIndex idx = index(path);
  dataChanged(idx,idx);
  layoutChanged();
}

void WcViewTreeItemModel::add( const sc::String& path, const svn::WcStatuss& statuss, bool deepStatus )
{
  // add to parent
  WcViewTreeItemFolder* parent = _m->getParentFolder(path);
  assert(parent);

  QModelIndex parentIdx;
  WcViewTreeItemFolder* grandparent = _m->getParentFolder(parent->path());
  if( grandparent )
  {
    int parentRow = grandparent->row(parent->path());
    parentIdx = createIndex( parentRow, 0, grandparent->child(parentRow) );
  }

  int row = parent->insertRow(path);

  beginInsertRows( parentIdx, row, row );
  
  WcViewTreeItem* item = new WcViewTreeItem(*statuss.begin());
  parent->insert( row, item );

  // always add childs container, even if we do not have any childs.
  WcViewTreeItemFolder* fitem = new WcViewTreeItemFolder(path,statuss,deepStatus);
  _m->index().insert( path, fitem );  
  
  // triggers filter check.
  endInsertRows();

  // no childs?
  if( statuss.size() <= 1 )
    return;

  // notify adding of childs
  QModelIndex idx = createIndex( row, 0, item );
  beginInsertRows( idx, 0, statuss.size()-2 /* 0 based & ignore folder */ );
  endInsertRows();
}

void WcViewTreeItemModel::remove( const sc::String& path, const svn::WcStatuss& statuss )
{
  //WcViewTreeItemFolder* item   = _m->getItemFolder(path);
  WcViewTreeItemFolder* parent = _m->getParentFolder(path);

  int row = parent->row(path);
  if( row == -1 )
    return;

#if 0 // crashes
  // remove childs from path if we know path.
  if( item )
  {
    QModelIndex idx;
    if( parent )
      idx = createIndex( parent->row(path), 0, parent->child(row) );

    // doesn't handle empty dirs, ie only a single child
    beginRemoveRows( idx, 0, item->childCount()-1-1 );
    delete _m->index().take(path);
    endRemoveRows();
  }
#endif

  // remove path from its parent.
  if( parent )
  {
    WcViewTreeItemFolder* grandparent = _m->getParentFolder(parent->path());

    QModelIndex parentIdx;
    if( grandparent )
    {
      int parentRow = grandparent->row(parent->path());
      parentIdx = createIndex( parentRow, 0, grandparent->child(parentRow) );
    }

    beginRemoveRows( parentIdx, row, row );
    parent->remove(row);

    if( _m->index().contains(path) )
      delete _m->index().take(path);

    endRemoveRows();
  }
}


#if 0
  WcViewTreeItemFolder

  WcViewTreeItems::const_iterator lowerBound( WcViewTreeItems::const_iterator begin, WcViewTreeItems::const_iterator end, const sc::String& path ) const
  {
    WcViewTreeItems::const_iterator middle;
    int n = end - begin;
    int half;

    while( n > 0 )
    {
      half   = n >> 1;
      middle = begin + half;

      if( (*middle)->path() < path )
      {
        begin = middle + 1;
        n    -= half + 1;
      }
      else
      {
        n = half;
      }
    }
    return begin;
  }

  WcViewTreeItems::const_iterator find( WcViewTreeItems::const_iterator begin, WcViewTreeItems::const_iterator end, const sc::String& path ) const
  {
    int l = 0;
    int r = end - begin - 1;
    if( r < 0 )
      return end;
    int i = (l + r + 1) / 2;

    while( r != l )
    {
      if( path < begin[i]->path() )
        r = i - 1;
      else
        l = i;
      i = (l + r + 1) / 2;
    }
    if ( begin[i]->path() < path || path < begin[i]->path() )
      return end;
    else
      return begin + i;
  }
#endif

